/*
 * fmt48: a HP48SX formatting/translation utility  V. 0.5.
 *
 * (c) 1990, 1991 Gadiel Seroussi
 *
 * May be copied, used, and modified in any way, but not for commercial
 * use, and provided that all the documentation and disclaimers are included
 * with the source.

DISCLAIMERS, WARRANTIES, GUARANTEES: the program is guaranteed to be 
bug-ridden, and to fail when you most need it. Other than that, no guarantees
or warranties, implicit or explicit.  I work for HP, but this was done on my 
free time, and the company has nothing to do with it.

 */


#include <stdio.h>
#include <string.h>
#include <malloc.h>


#define   VERSION "V. 0.5"
int debug = 0;

#define	INDT	2
#define NLINE	1
#define DELAYED 2

#define IN_LINELEN	10000
#define MAX_TOKEN	(IN_LINELEN+16)
#define OUT_LINELEN	50
#define MAX_INDENT	((3*OUT_LINELEN)/4)
#define MAX_PATH	1024

#define ROOTDIR		""

#define PDL_COMMENT_CHAR    '/'
#define PDL_COMMENT_ALT     '\377'
#define LINE_COMMENT_ALT    '\376'

#define COMMENT_INDENT	52

#define is_comment(x) ( ((x)==COMMENT)||((x)==LINE_COMMENT)||\
                        ((x)==PDL_COMMENT)||((x)==MODES_COMMENT) )

#define TBD_MODE 	1
#define	DIR_MODE	2
#define PACKED_MODE	3

enum	token_type { OTHER, 
		     LPGM, RPGM, LCURL, RCURL, 
		     LOCAL_DVAR,
		     DQUOT, LDQUOT, RDQUOT,
		     SQUOT, NAME, LSQUOT, RSQUOT,
		     CSTRING,
		     LBRKT, RBRKT, 
		     DLBRKT, DRBRKT, DIR, 
		     IF, IFERR, THEN, ELSE, END, CASE,
		     WHILE, REPEAT,
		     DO, UNTIL,
		     FOR, STEP, NEXT,
		     START,
		     SHARP,
		     GROB,
		     DIVIDE,
		     COMMENT,
		     LINE_COMMENT,
		     MODES_COMMENT,
		     PDL_COMMENT,
		     LINKED, ARRAY,
		     LONG, REAL, COMPLEX,
		     LIBRARY, DATA,
		     EOL,
		     LAST_TOKEN
		   };

typedef struct {
	char *str;
	enum token_type type;
	int  indent_before,
	     indent_after,
	     nl_before,
	     nl_after,
	     nest_index, /* index for nesting counter */
	     nest_delta;       /* nesting increment for above counter */
} token_node;

token_node   token_table[] =  {

/* Token	Type		ind_bef ind_af  nl_bef  nl_aft indx delta */

{"\\<<", 	LPGM ,		0,	1,	NLINE,	NLINE,RPGM,  1},
{"\\>>", 	RPGM ,		-1,	0,	NLINE,	NLINE,RPGM, -1},
{"{",    	LCURL,		0,	0,	0,	0,    RCURL, 1},
{"}",		RCURL,		0,	0,	0,	0,    RCURL,-1},
{"'dvar",	LOCAL_DVAR,	0,	0,	0,	0,    LOCAL_DVAR, 0},
{"\"",		DQUOT,		0,	0,	0,	0,    DQUOT, 1},
{"\"",		LDQUOT,	 	0,	0,	0,	0,    DQUOT, 1},
{"\"",		RDQUOT,	 	0,	0,	0,	0,    DQUOT,-1},
{"'",		SQUOT,		0,	0,	0,	0,    SQUOT, 1},
{"'",		LSQUOT,		0,	0,	0,	0,    SQUOT, 1},
{"'",		RSQUOT,		0,	0,	0,	0,    SQUOT,-1},
{"'",		NAME,		0,	0,	0,	0,    NAME, 0},
{"C$",		CSTRING,	0,	0,	0,	0,    GROB,  2},
{"[",		LBRKT,		0,	0,	0,	0,    RBRKT, 1},
{"]",		RBRKT,		0,	0,	0,	0,    RBRKT,-1},
{"[[",		DLBRKT,		0,	0,	0,	0,    RBRKT, 2},
{"]]",		DRBRKT,		0,	0,	0,	0,    RBRKT,-2},
{"DIR",		DIR  ,		0,	1,	NLINE,	NLINE,END,   1},
{"IF",		IF   ,		0,	0,	NLINE,	0,    END,   0},
{"IFERR",	IFERR,		0,	0,	NLINE,	NLINE,END,   0},
{"THEN",	THEN ,		0,	1,	NLINE,	NLINE,END,   1},
{"ELSE",	ELSE ,		-1,	1,	NLINE,	NLINE,OTHER, 0},
{"END",		END  ,		-1,	0,	NLINE,	NLINE,END,  -1},
{"CASE",	CASE ,		0,	1,	NLINE,	NLINE,END,   1},
{"WHILE",	WHILE,		0,	0,	NLINE,	0,    END,   1},
{"REPEAT",	REPEAT,		0,	1,	NLINE,	NLINE,END,   0},
{"DO",		DO   ,		0,	1,	NLINE,	NLINE,END,   1},
{"UNTIL",	UNTIL,		-1,	1,	NLINE,	0,    END,   0},
{"FOR",		FOR  ,		0,	1,	NLINE,	DELAYED,NEXT,  1},
{"STEP",	STEP ,		-1,	0,	NLINE,	NLINE,NEXT, -1},
{"NEXT",	NEXT ,		-1,	0,	NLINE,	NLINE,NEXT, -1},
{"START",	START,		0,	1,	NLINE,	NLINE,NEXT,  1},
{"GROB",	GROB,		0,	0,	0,	0,    GROB,  3},
{"/",		DIVIDE,		0,	0,	0,	0,    DIVIDE,  0},
{"Linked",	LINKED,		0,	0,	0,	0,    ARRAY, 1},
{"Array",	ARRAY,		0,	0,	0,	0,    ARRAY, -1},
{"Long",	LONG,		0,	0,	0,	0,    LONG,  1},
{"Real",	REAL,		0,	0,	0,	0,    LONG, -1},
{"Complex",	COMPLEX,	0,	0,	0,	0,    LONG, -1},
{"Library",	LIBRARY,	0,	0,	0,	0,    DATA,  1},
{"Data",	DATA,		0,	0,	0,	0,    DATA, -1},
{"@",		COMMENT,	0,	0,	0,	NLINE,OTHER, 0},
{"@",		LINE_COMMENT,	0,	0,	NLINE,	NLINE,OTHER, 0},
{"%%HP:",	MODES_COMMENT,	0,	0,	NLINE,	NLINE,OTHER, 0},
{"/",		PDL_COMMENT,	0,	0,	NLINE,	NLINE,OTHER, 0},
{"#",		SHARP,		0,	0,	0,	0,    GROB,  1},
{"\n",		EOL  ,		0,	0,	0,	0,    OTHER, 0},
{"",		OTHER,		0,	0,	0,	0,    OTHER, 0},
{NULL,		LAST_TOKEN,		}
};

/* cross-reference for tokens */
token_node   *token_xref[LAST_TOKEN+1];

/* set of nesting counters */
struct nest_counters {
	int	counter[LAST_TOKEN+1];
	int	total_nesting;
};

/* Function declarations: */
char *in_token() , *get_line();
token_node *find_token();

/* Node for path link list */
struct path_node {
	char	*name;
	struct  path_node *next;
};

/* head for path link list: */
struct path_node *path;

/* root directory name */
char   *rootdir = ROOTDIR;

/* input line count (global) */
int inp_linecount = 0;

/* operation modes */
int	app_mode = 0, /* "Application" mode (input is a PDL application) */
	transl_chars = 0, /* do not translate special characters */
	transl_mode = 0, /* translation mode, default: NONE */
        name_mode = 0, /* if set to 1, names are are written in separate
	                  lines */
	quiet = 0,    /* if 1, do not add comments in directories */
	comment_indent = COMMENT_INDENT,
	line_indent = INDT,
	out_linelen = OUT_LINELEN;

int	global_nesting = 0;  /* global object nesting level */

char outfilename[256] = "";

main(argc, argv)
int argc;
char *argv[];
{
	FILE *fin ,
	     *fout = stdout;
	char *infilename = NULL,
	     *file_suffix = "";
	int   i, obj_count;

	for ( i=1; i<argc; i++ ) {
		if ( argv[i][0] == '-' ) {
			switch ( argv[i][1] ) {
				case 'd':
					rootdir = argv[i]+2;
					break;
				case 'T':
					transl_chars = 1;
					break;
				case 't':
					/* set translation mode ON, whether it's
					   DIR or PACKED will be determined
					   later */
					transl_mode = TBD_MODE; 
					break;
				case 'a':
					/* set "application" mode 
					   for PDL files */
					app_mode = 1;
					break;
				case 'n':
					/* put names in separate lines
					   for PDL translation files */
					name_mode = 1;
					break;
				case 'q':
					quiet = 1;
					break;

				case 'l':
					if ( sscanf(argv[i],"-l%d",&out_linelen) != 1 ) {
						usage();
						exit(10);
					}
					break;

				case 'i':
					if ( sscanf(argv[i],"-i%d",&line_indent) != 1 ) {
						usage();
						exit(10);
					}
					break;

				case 'I':
					if ( sscanf(argv[i],"-I%d",&comment_indent) != 1 ) {
						usage();
						exit(10);
					}
					break;

				case 'D':
					debug = 1;
					break;
				
				default:
					usage();
					exit(0);
					break;
			}
		}
		else {
			if ( infilename == NULL ) 
				infilename = argv[i];
			else
				strncpy(outfilename, argv[i], 256);
		}
	}

	file_suffix = transl_mode ? ".app" : ".f48";
	fprintf(stderr,"Translation mode is %s.\n",
		transl_mode ? "ON" : "NONE");
		              
	if ( infilename == NULL ) {
		fin = stdin;
		fout = stdout;
	}
	else { 
		if (  (fin=fopen(infilename,"r"))== NULL ) {
			fprintf(stderr,"cannot open input file %s\n",infilename);
			exit(10);
	
		}
		if ( strlen(outfilename) == 0 ) {
			char *p;
		        /* take suffix from infilename */
			strcpy(outfilename, infilename);
			if (  ( p = strrchr(outfilename,'.' ) ) != NULL )
				*p = 0;
			strcat(outfilename,file_suffix);
		}
		if ( (fout=fopen(outfilename,"w"))==NULL) {
			fprintf(stderr,"cannot open output file %s\n",outfilename);
			exit(10);
		}
		fprintf(stderr,"writing to output file %s\n",outfilename);
	}

	init_tables();
	for ( obj_count=0; get_object(fin, fout) == 0; obj_count++ );
	fprintf(stderr,"@  %d objects read\n",obj_count);
	exit(0);
}


init_tables()
{
	token_node *p = token_table - 1;

        
	/* build cross-reference for tokens */
	do {
		p ++;
		token_xref[(int)(p->type)] = p;
	} while ( p->type != LAST_TOKEN );

	/* if name_mode==1 , set nl_before and nl_after to one
	   for names */
	if ( name_mode ) {
		p = token_xref[NAME];
		p->nl_before = p->nl_after = 1;
	}
}

get_object(fin, fout) 
FILE *fin, *fout;
{
	static char  last_name[MAX_TOKEN+1], /* remember last name read  
					       before DIR              */
                     path_string[MAX_PATH];  /* string for path */
	char *tok;
	token_node   cur_token, *tkn = &cur_token;
	int  empty_token = 0;
	struct nest_counters nestc;  /* nesting counters for this object */
	int i;

	

	for ( i=0; i<= LAST_TOKEN; i++ )
		nestc.counter[i] = 0;
	nestc.total_nesting = 0;

	global_nesting ++;

	do {
		if ( get_token(fin, tkn) == EOF ) {
			flush_fout(fout);  /* flush output */
			global_nesting --;
			return -1;
		}
		tok = tkn->str;
		
#ifdef DEBUG
		if ( debug ) {
		    printf("type = %d  token = %s \n",tkn->type, tok);
		    fflush(stdout);
		}
#endif
		if ( tkn->type != EOL ) {
		    if ( transl_mode == PACKED_MODE && tkn->type == NAME &&
		         global_nesting == 1 && nestc.total_nesting == 1 ) {
			char *p;
	            /* a new variable in a PACKED archive */
			flush_fout(fout);  /* flush output */
			fprintf(fout,"//\n/Variable ");
			/* print name (without the quotes) */
			for ( p = tok+1; *p != '\''; p++ )
				putc(*p, fout);
			fprintf(fout,"\n");
		    }
		    else  { /* check if this is the outer DIR in a DIR file
		             being translated to `.app', or the outer << >>
			     in a PACKED file being translated to `.app' .
			     The outer END in a DIR file is handled separately.
			    */
			if ( 
			     transl_mode && global_nesting == 1 && 
			        ( nestc.total_nesting == 0 &&
			          ( tkn->type == DIR || tkn->type == LPGM ) ||
   				  nestc.total_nesting == 1 && tkn->type == RPGM 
				) 
			    )  {
				flush_fout(fout);  /* do not output token */
				if ( tkn->type == DIR ) {
				/* set translation mode */
					transl_mode = DIR_MODE;
					fprintf(stderr,"Input file is of DIR type.\n");
				}
				else if ( tkn->type == LPGM ) {
					transl_mode = PACKED_MODE;
					fprintf(stderr,"Input file is of PACKED type.\n");
				}
			}
			else
		    /* nothing special about this token, print it */
				out_token(fout, tkn);
		    }
		}

		empty_token = 0;

		if ( tkn->type == DIR ) {
		/* process a new directory */
			update_nesting(tkn, &nestc);
			/* add dir name (if any) to path */
			if ( path == NULL )
				strncpy(last_name, rootdir, MAX_TOKEN);
			add_path(&path, last_name);

			do {
				/* go to new output line */
				flush_fout(fout);

				/* look for a variable name */
				if ( get_token(fin, tkn) == EOF )
				    break;
				
				switch ( tkn->type ) {
				case LINE_COMMENT:
				case PDL_COMMENT:
				case MODES_COMMENT:
				case COMMENT:
					out_token(fout, tkn);
					continue;

				case EOL:
					continue;

				case OTHER:   /* a name will fall here */
				/* we'll allow the following as names, but it's 
				   flakey:                                    */
				case DATA:
				case LONG:
				case REAL:
				case COMPLEX:
					/* remember the var name */
					strncpy(last_name, tkn->str, MAX_TOKEN);

					/* generate the path for printing: */
					gen_path(path_string, path);

					/* print the var name */
					if ( transl_mode == DIR_MODE ) {
						/* DIR->APP translation */
						/* flush output */
																flush_fout(fout);
						fprintf(fout,"/Variable ");
						if ( strlen(path_string) )
							fprintf(fout,"{ %s %s }\n",path_string,tkn->str);
						else 
							fprintf(fout,"%s\n",tkn->str);
					}
					else
						out_token(fout, tkn);
					/* print a comment with then name
					   and path */
					if ( ! quiet ) {
					    fprintf(fout, "@\n@ Variable: %s\n",
						     tkn->str);
					    fprintf(fout,
					"@\n@ Path: { %s}\n@\n",path_string);
					}

					/* get an object */
					get_object(fin, fout);
					break;

				case END:
					/* end of directory */
					/* check if this is the outer END in
					   a DIR file being translated to 
					   `.app'.                           */
					if ( transl_mode == DIR_MODE &&
					     global_nesting == 1 &&
					     nestc.total_nesting == 1
					   )
						flush_fout(fout);
					else
						out_token(fout, tkn);
					update_nesting(tkn, &nestc);
					/* drop last node from path */
					remove_path(&path);
					break;

				default: /* wrong type of token */
					fprintf(stderr,"DIR: expected name, got %s (line %d)\n", tkn->str, inp_linecount);
					continue;
				}
			} while ( tkn->type != END );
		}	

		else switch ( tkn->type ) {

			case OTHER: /* names, numbers, etc. */
				if ( nestc.counter[GROB] > 0 ) {
					nestc.counter[GROB] --;
					nestc.total_nesting --;
				}
				break;


			case DQUOT:
			case SQUOT:   /* "open" and "close" are the same */
				if ( nestc.counter[tkn->nest_index] > 0 ) {
					nestc.counter[tkn->nest_index] --;
					nestc.total_nesting --;
				}
				else {
					nestc.counter[tkn->nest_index] ++;
					nestc.total_nesting ++;
				}
				break;

			case EOL:
			case COMMENT:
			case PDL_COMMENT:
			case LINE_COMMENT:
			case MODES_COMMENT:
				empty_token = 1;
				break;

			default:
				update_nesting( tkn, &nestc);
				break;

		}

	} while ( nestc.total_nesting > 0 || empty_token );

	global_nesting --;
	return 0;
}

update_nesting( tkn, nestp)
token_node *tkn;
struct nest_counters *nestp;
{
	int ctr;

	ctr =  ( nestp->counter[ tkn->nest_index ] += tkn->nest_delta );
	nestp->total_nesting += tkn->nest_delta;
	if ( debug ) {
		printf("Token=<%s> Incr.=%d index=%d nestp=%4x New ctr.=%d Total nest=%d\n",
			tkn->str, tkn->nest_delta, tkn->nest_index, nestp, ctr,
			nestp->total_nesting);
	}
	if ( ctr < 0 ) {
		fprintf(stderr,"Negative nesting counter for token %s (line %d)\n",
			   tkn->str, inp_linecount);
		return -1;
	}
	if ( nestp->total_nesting < 0 ) {
		fprintf(stderr,"Negative total nesting counter for token %s (line %d)\n",
			   tkn->str, inp_linecount);
		return -1;
	}
}

gen_path(path_string, path)
char *path_string;
struct path_node *path;
{
	struct path_node *p;
	int	len;

        path_string[0] = 0;
	len = 0;
	for ( p=path; p != NULL; p = p->next ) {
		if ( *(p->name) && ( len + strlen(p->name) < MAX_PATH ) ) {
			strcat(path_string, p->name);
			strcat(path_string," ");
		}
	}
}

add_path(pathp, name)
struct path_node **pathp;
char *name;
{
	struct path_node *p, *p2, *path = *pathp;
	int    len = strlen(name);

	if ( path != NULL  )    /* look for last node in list */
	    for ( p=path; p->next != NULL; p = p->next );

	if ( (p2 = (struct path_node *)malloc( sizeof(*p))) == NULL ) {
		fprintf(stderr,"add_path: could not allocate path node\n");
		exit(10);
	}
	p2->next = NULL;
	if ( ( p2->name = malloc(len+1) ) == NULL ) {
		fprintf(stderr,"add_path: could not allocate path name\n");
		exit(10);
	}
	strcpy(p2->name, name);
	if ( path != NULL )
		p->next = p2;
	else
		*pathp = p2;
}

remove_path(pathp)
struct path_node **pathp;
{
/* unlink last node from path */
	struct path_node *p, *p2, *path = *pathp;

	if ( path == NULL ) {
		fprintf(stderr,"trying to remove from an empty path\n");
		return -1;
	}
	for ( p=path, p2=NULL; p->next != NULL; p2 = p, p = p->next );

	free(p->name);
	free(p);
	if ( p2 )
	    p2->next = NULL;
	else
	    *pathp = NULL;   /* emptied the whole list */
	return 0;
}

int
get_token(fp, tkn)
FILE *fp;
token_node *tkn;
{

	token_node *tkn2;
	char    *tok, *read_token();

	if ( (tok = read_token(fp)) == NULL )    /* read from input file */
		return EOF;
	
	/* find pointer to token table */
	tkn2 = find_token( tok, token_table ); 

	*tkn = *tkn2;    /* copy token */
	tkn->str = tok;
	return 0;
	
}

char *
read_token(fp)
FILE *fp;
{

/* Read next token from input file. Handles quotes and comments,
   returning them as whole tokens.                            */

	char   *next_sep(), *next_nonsep();
	static char  inp_line[IN_LINELEN];
	static char  token[MAX_TOKEN+1]; /* used only in multi-line tokens */
	static char  *lp;     /* pointer to current position in input line */
		     
	static int   newline = 1;
	char         *tok, *lp2;
	

	if ( newline ) {    /* read new line */
		if ( get_line(inp_line,fp) == NULL )
			return NULL;
		inp_linecount ++;
		newline = 0;
		lp = inp_line;

		/* detect full-line comments and PDL directives: */
		if ( 
		     ( inp_line[0] == '@' ) || 
		     ( strncmp(inp_line, "%%HP", 4) == 0 )  ||
		     ( app_mode && ( inp_line[0] == PDL_COMMENT_CHAR ) )) 
		{
			/* it's a comment or PDL directive*/
			tok = inp_line; /* take whole line */
			tok[strlen(tok)-1] = '\0'; /* strip \n */

			switch ( inp_line[0] ) {
			case PDL_COMMENT_CHAR:
				/* replace first character of PDL directive,
				   to distinguish token from a division 
				   operator */
				inp_line[0] = PDL_COMMENT_ALT;
				break;
			case '@':
				/* distinguish full line comments from
				   regular comments */
				inp_line[0] = LINE_COMMENT_ALT;
				break;
			}

			newline = 1;    /* read a new line next time */
			return tok;
		}
	}
	
	lp = next_nonsep(lp);   /* find next non-separator */

	if  ( *lp == '\0' ) {  /* end of line */
		tok = "\n";
		newline = 1;
		return tok;
	}

	if ( *lp == '@' ) { /* take rest of line as comment */
		tok = lp;
		newline = 1;
		tok[strlen(tok)-1] = '\0'; /* strip \n */
		return tok;
	}

	if ( *lp == '"' ) {     /* quoted string: return as one token */
		strcpy(token, "\"");
		lp++;
		while ( (lp2 = strchr(lp, '"')) == NULL ) {
			strncat(token, lp, MAX_TOKEN-strlen(token));
			if ( get_line(inp_line,fp) == NULL ) {
				newline = 1;
				return token;   /* unfinished string! */
			}
			inp_linecount ++;
			lp = inp_line;
		}

		if ( *(lp2+1) ) 
		    * ++ lp2 = 0;

		strncat(token, lp, MAX_TOKEN-strlen(token));
		lp = lp2 + 1;
		return token;
	}


	lp2 = next_sep(lp);     /* find next separator (including '\0') */

	if ( *lp2 )
		*lp2 ++  = 0;
	tok = lp;
	lp = lp2;

	return tok;


}

static char 
	*separators = " \t\n\r";

char *
next_sep(lp)
char *lp;
{
	char *lp2;
	lp2 = strpbrk(lp, separators);
	if ( lp2 == NULL )
		return lp+strlen(lp);
	else
		return lp2;
}

char *
next_nonsep(lp)
char *lp;
{
	for ( ; *lp && (strchr(separators, *lp) != NULL); lp++ );
	return lp;
}

		
out_token( fp, tkn)
FILE	   *fp;
token_node *tkn;
{
	static	char	out_line[MAX_TOKEN+1] = "";
	static	int	indt_level = 0,
			newline = 1,
			delayed_after = 0;
	int     do_flush = 0;
	char	*tok = tkn->str;
	int	tklen;

	indt_level += line_indent *  tkn->indent_before;
	tklen = strlen(tok);

	if ( tkn->nl_before || tkn->type == EOL || 
			       tkn->type != COMMENT && 
			       strlen(out_line)+tklen >= out_linelen )  {
		flush_line( out_line, indt_level, newline, fp);
		newline = 1;
	}

	if ( tkn->type != EOL ) {
	    if (  newline && tklen+strlen(out_line) > out_linelen  
		  || tkn->type == LINE_COMMENT 
		  || tkn->type == MODES_COMMENT
		  || tkn->type == PDL_COMMENT ) {
		/* do not indent */
		strcpy(out_line, tok);
		do_flush = 1;  /* make sure line is flushed */
	    }
	    else {
		if ( tkn->type == COMMENT ) {
			int i;
			if ( strlen(out_line) >= comment_indent )  {
				/* flush the line */
				flush_line( out_line, indt_level, newline, fp );
			}
			for ( i=0; strlen(out_line) < comment_indent-1; i++ )
				strcat(out_line," ");
		}
		if ( tkn->type == DIVIDE && strlen(out_line) == 0 )
		/* make sure that the divide operator "/" never
		   comes at the beginning of a line */
		   strcat( out_line, " ");

		strcat(out_line,tok);
		if ( strlen(out_line) < out_linelen )
		    strcat(out_line," ");
	    }
	    newline = 0;
	}

	indt_level += line_indent * tkn->indent_after;

	if ( delayed_after || (tkn->nl_after == NLINE) || do_flush ) {
		flush_line(out_line, indt_level, 0, fp);
		newline = 1;
		delayed_after = 0;
		do_flush = 0;
	}
	else if ( tkn->nl_after == DELAYED ) {
		delayed_after = 1;
	}
}

flush_fout(fout)
FILE *fout;
{
	out_token( fout, token_xref[EOL] );
}

flush_line(line, indt_level, init_only, fp)
char *line;
int  indt_level, init_only;
FILE *fp;
{
	int i;

	if ( !init_only )
		fprintf(fp,"%s\n",line);
	fflush(fp);
	for ( i=0; i<indt_level && i < MAX_INDENT; i++ )
		line[i] = ' ';
	line[i] = 0;
}
	
	
token_node *
find_token( token, token_table)
char *token;
token_node *token_table;
{
	token_node *tkp;
	int        tklen;
	char	   first, last;

	tklen = strlen(token);
	first = token[0];
	last  = tklen ? token[tklen-1] : '\0';

	/* handle comments: */
	if ( 
	     first == '@' || 
	     first == LINE_COMMENT_ALT ||
	     strncmp(token, "%%HP", 4) == 0  ||
	     app_mode && ( first == PDL_COMMENT_ALT ) ) 
	{
		switch( first ) {
		case PDL_COMMENT_ALT:
		    token[0] = PDL_COMMENT_CHAR;   /* restore first char */
		    return token_xref[PDL_COMMENT];

		case LINE_COMMENT_ALT:
		    token[0] = '@';   /* restore first char */
		    return token_xref[LINE_COMMENT];

		case '@':
		    return token_xref[COMMENT];
		
		case '%':
		    return token_xref[MODES_COMMENT];
		
		}

	}

	/* Tokens from the table: */
	for ( tkp = token_table; tkp->type != OTHER; tkp++ )
		if ( strcmp(tkp->str, token) == 0 )
			return tkp;

	/* quotes are treated separately, as they may be 
	   "glued" to their arguments, but still need to be
	   recognized.
	*/
	if ( first == '"' && tklen > 1 ) 
		if ( last == '"' )
			return token_xref[OTHER];
		else
			return token_xref[LDQUOT];

	if ( last == '"' && tklen > 1 )
		return token_xref[RDQUOT];
	
	if ( first == '\'' && tklen > 1 ) 
		if ( last == '\'' )
			return token_xref[NAME];
		else
			return token_xref[LSQUOT];
	if ( last == '\'' && tklen > 1 )
		return token_xref[RSQUOT];
	

	
	return tkp;
}

char *
get_line( inp_line, fp )
char *inp_line;
FILE *fp;
{
	static char 	tmpline[IN_LINELEN+2];

	inp_line[0] = 0;
	if ( fgets(tmpline,IN_LINELEN,fp) == NULL )
		return NULL;

	if ( transl_chars ) 
		translate_line(tmpline, inp_line, IN_LINELEN );
	else
		strncpy(inp_line, tmpline, IN_LINELEN);
	return inp_line;
}

translate_line(inpline, outline, len)
unsigned char *inpline, *outline;
int  len;
{
	unsigned char *p, escseq[5];
	int      l = 0;

	len --;
	outline[0] = 0;
	for ( p = inpline; *p ; p++ ) {
		if (  
			( *p < 32 && *p != 10 && *p != 13 ) ||
			( *p >= 127 ) ) {
			sprintf(escseq,"\\%03u",(unsigned)(*p));
			if ( l + 4 > len ) {
				fprintf(stderr,"WARNING: Line truncated\n");
				return;
			}
			strcat(outline,escseq);
			l += 4;
		} 
		else
		{
			if ( l + 1 > len ) {
				fprintf(stderr,"WARNING: Line truncated\n");
				return;
			}
			outline[l++] = *p;
		}
	}
	outline[l] = 0;
}

			

usage()
{
	fprintf(stderr,"FMT48  %s\n\n",VERSION);
	fprintf(stderr,
"usage: fmt48 [ flags ] [ filein ] [ fileout ]\n");
	fprintf(stderr,
"\nFlags:\n\
-a     :  Format PDL `.app' file.\n\
-t     :  Translate HP48 DIR or `packed' file to PDL `.app' file.\n\
-n     :  Put quoted names on separate lines (for `transapp').\n\
-d\"path\": Set root directory to `path'.\n\
-T     :  Translate characters 0-31 (except 10, 13) and 127-255 to\n\
          escape sequences \\ddd.\n");

	fprintf(stderr,"\
-l<num>:  Set output line length (w/o comments) to <num> (default = %d).\n\
-i<num>:  Set indent increment to <num> (default = %d).\n\
-I<num>:  Set comment indent to <num> (default = %d).\n\
-q     :  Quiet mode: do not add comments to DIR files.\n",
          OUT_LINELEN, INDT, COMMENT_INDENT);
	fprintf(stderr,
"\nFiles:\n\
filein : input file (default = stdin).\n\
fileout: output file. Defaults:\n\
         If filein=stdin then fileout=stdout,\n\
	 if filein=name.ext, and -t was not specified, then fileout=name.f48,\n\
	 if filein=name.ext, and -t was specified, then fileout=name.app.\n");
}
